# -*- coding: utf-8 -*-

"""WebHelpers used in tgapp-categories."""


def images_with_image_name(category, image_name):
    return [c for c in category.images if c.image_name == image_name]


def content_of_first_small_image(category):
    try:
        return images_with_image_name(category, 'image_small')[0].content
    except IndexError:
        return None


def content_of_first_big_image(category):
    try:
        return images_with_image_name(category, 'image_big')[0].content
    except IndexError:
        return None


from tgext.pluggable.utils import instance_primary_key



from ming import schema as s
from ming.odm import FieldProperty, RelationProperty, ForeignIdProperty
from ming.odm.declarative import MappedClass
from bson import ObjectId
import re

from tgappcategories.model import DBSession

from depot.fields.ming import UploadedFileProperty


class Category(MappedClass):
    class __mongometa__:
        session = DBSession
        name = 'tgappcategories_categories'
        indexes = [
            ('name',),
            ('path',),
        ]

    _id = FieldProperty(s.ObjectId)
    
    name = FieldProperty(s.String)
    description = FieldProperty(s.String)

    images = RelationProperty('CategoryImage')

    path = FieldProperty(s.String)
    depth = FieldProperty(s.Int)

    @property
    def descendants(self):
        path = '~%s' % self._id if not self.path else self.path
        rgx = re.compile('^%s~*' % path)
        return Category.query.find({'path': rgx, '_id': {'$ne': self._id}}).all()

    @property
    def children(self):
        next_depth = self.depth + 1
        path = '~%s' % self._id if self.path is None else self.path
        rgx = re.compile('^%s~*' % path)
        return Category.query.find({'path': rgx, 'depth': next_depth}).all()

    @property
    def parent(self):
        try:
            return Category.query.find({'_id': ObjectId(self.path.split('~')[-1])}).first()
        except:
            return None

    @property
    def siblings(self):
        return Category.query.find({'path': self.path, '_id': {'$ne': self._id}}).all()  


class CategoryImage(MappedClass):
    class __mongometa__:
        session = DBSession
        name = 'tgappcategories_images'

    _id = FieldProperty(s.ObjectId)

    content = UploadedFileProperty(upload_storage='category_image')

    image_name = FieldProperty(s.String)

    category_id = ForeignIdProperty('Category')
    category = RelationProperty('Category')



# -*- coding: utf-8 -*-
"""Main Controller"""

import tg
from tg import TGController
from tg import expose, flash, url, redirect, validate, predicates, abort
from tg.i18n import ugettext as _

from tgappcategories import model
from tgext.pluggable import plug_url

from tgappcategories.lib import get_new_category_form, get_edit_category_form


class RootController(TGController):
    allow_only = predicates.has_permission('tgappcategories')

    @expose('tgappcategories.templates.index')
    def index(self):
        categories = model.provider.query(model.Category, order_by='path')
        return dict(categories_count=categories[0],
                    categories=categories[1],
                    mount_point=self.mount_point,
                    )

    @expose('tgappcategories.templates.new_category')
    def new_category(self, **_):
        return dict(form=get_new_category_form(),
                    mount_point=self.mount_point,
                    action=plug_url('tgappcategories', '/create_category'),
                    values=None,
                    )

    @expose()
    @validate(get_new_category_form(), error_handler=new_category)
    def create_category(self, **kwargs):
        parent = model.provider.get_obj(model.Category, {'_id': kwargs.get('pare\
nt_id')})
        image_small = {
            'content': kwargs.get('image_small'),
            'image_name': 'image_small',
        }
        img_small = model.provider.create(model.CategoryImage, image_small)
        image_big = {
            'content': kwargs.get('image_big'),
            'image_name': 'image_big',
        }
        img_big = model.provider.create(model.CategoryImage, image_big)
        category = {
            'name': kwargs.get('name'),
            'description': kwargs.get('description'),
            'images': [img_small, img_big],
            'path': parent.path + '~' + str(parent._id) if parent else '',
            'depth': parent.depth + 1 if parent else 1,
        }
        model.provider.create(model.Category, category)
        flash(_('Category created.'))
        return redirect(url(self.mount_point))

    @expose('tgappcategories.templates.edit_category')
    def edit_category(self, category_id, **_):
        category = model.provider.get_obj(model.Category, {'_id': category_id}) or abort(404)
        category.image_small_id = category.images[0]._id
        category.image_big_id = category.images[1]._id
        category.parent_id = category.path.split('~')[-1]
        return dict(form=get_edit_category_form(),
                    mount_point=self.mount_point,
                    action=plug_url('tgappcategories', '/update_category/' + category_id),
                    values=category,
                    )

    @expose()
    @validate(get_edit_category_form(), error_handler=edit_category)
    def update_category(self, category_id, **kwargs):
        category = model.provider.get_obj(model.Category, {'_id': category_id})
        category.name = kwargs.get('name')
        category.description = kwargs.get('description')

        img_small = img_big = None
        if kwargs.get('image_small') is not None:
            image_small = {
                'content': kwargs.get('image_small'),
                'image_name': 'image_small',
            }
            img_small = model.provider.create(model.CategoryImage, image_small)
        if kwargs.get('image_big') is not None:
            image_big = {
                'content': kwargs.get('image_big'),
                'image_name': 'image_big',
            }
            img_big = model.provider.create(model.CategoryImage, image_big)
        category_parent_id = category.path.split('~')[-1]
        if kwargs.get('parent_id') != category_parent_id:
            old_path = category.path
            old_depth = category.depth
            new_parent = model.provider.get_obj(model.Category, {'_id': kwargs.get('parent_id')})
            new_path = (new_parent.path if new_parent else '') \
                + ('~' + str(new_parent._id) if new_parent else '')
            if new_parent and new_parent._id == category._id:
                abort(412, _('Cannot move category into itself'))
            new_depth = new_parent.depth + 1 if new_parent else 1
            for d in category.descendants:
                if new_path == d.path + '~' + str(d._id):
                    abort(412, _('Cannot move category to a descendant category'))
                # calculating descendant depth
                d.depth += (new_depth - old_depth)
                # calculate descendant path
                if old_path != '':
                    d.path = d.path.replace(old_path, new_path)
                else:
                    d.path = new_path + d.path
            category.path = new_path
            category.depth = new_depth
        original_small = model.provider.get_obj(model.CategoryImage,
                                                {'_id': kwargs.get('image_small_id')})
        original_big = model.provider.get_obj(model.CategoryImage,
                                              {'_id': kwargs.get('image_big_id')})
        category.images = [img_small or original_small, img_big or original_big]

            
        tg.hooks.notify('categories.after_update', args=(category, kwargs))
        flash(_('Category updated.'))
        return redirect(url(self.mount_point))

    @expose()
    def delete_category(self, category_id):
        category = model.provider.get_obj(model.Category, dict(_id=category_id)) or redirect(url(self.mount_point))
        if len(category.descendants) == 0:
            model.provider.delete(model.Category, dict(_id=category_id))
            flash(_('Category deleted'))
            return redirect(url(self.mount_point))
        else:
            abort(412, _('Cannot delete category because it has descendants'))



# -*- coding: utf-8 -*-

from tg import config


def get_new_category_form():
    new_category_config = config['_pluggable_tgappcategories_config']
    new_category_form = new_category_config.get('new_category_form_instance')
    if not new_category_form:
        form_path = new_category_config.get('new_category_form', 'tgappcategories.lib.forms.NewCategory')
        module, form_name = form_path.rsplit('.', 1)
        module = __import__(module, fromlist=form_name)
        form_class = getattr(module, form_name)
        new_category_form = new_category_config['new_category_form_instance'] = form_class()

    return new_category_form


def get_edit_category_form():
    edit_category_config = config['_pluggable_tgappcategories_config']
    edit_category_form = edit_category_config.get('edit_category_form_instance')
    if not edit_category_form:
        form_path = edit_category_config.get('edit_category_form', 'tgappcategories.lib.forms.EditCategory')
        module, form_name = form_path.rsplit('.', 1)
        module = __import__(module, fromlist=form_name)
        form_class = getattr(module, form_name)
        edit_category_form = edit_category_config['edit_category_form_instance'] = form_class()
    return edit_category_form



from tg import TGController, expose, request


class RootController(TGController):
    def __call__(self, *args, **kwargs):
        request.identity = request.environ.get('repoze.who.identity')
        return super(RootController, self).__call__(*args, **kwargs)

    @expose()
    def index(self):
        return 'HELLO'


# -*- coding: utf-8 -*-
import logging
import tg
from tgext.pluggable import PluggableSession

log = logging.getLogger('tgappcategories')

DBSession = PluggableSession()
provider = None

Category = None
CategoryImage = None


def init_model(app_session):
    DBSession.configure(app_session)


def configure_models():
    global provider, Category, CategoryImage

    if tg.config.get('use_sqlalchemy', False):
        log.info('Configuring TgappCategories for SQLAlchemy')
        from tgappcategories.model.sqla.models import Category, CategoryImage
        from sprox.sa.provider import SAORMProvider
        provider = SAORMProvider(session=DBSession, engine=False)
    elif tg.config.get('use_ming', False):
        log.info('Configuring TgappCategories for Ming')
        from tgappcategories.model.ming.models import Category, CategoryImage
        from sprox.mg.provider import MingProvider
        provider = MingProvider(DBSession)
    else:
        raise ValueError('TgappCategories should be used with sqlalchemy or ming')



from sqlalchemy import Column, ForeignKey, event
from sqlalchemy.types import Unicode, Integer

from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship
from sqlalchemy.orm.exc import NoResultFound

from depot.fields.sqlalchemy import UploadedFileField
from tgext.pluggable import app_model


DeclarativeBase = declarative_base()


class Category(DeclarativeBase):
    __tablename__ = 'tgappcategories_categories'

    _id = Column(Integer, autoincrement=True, primary_key=True)

    name = Column(Unicode(255))
    description = Column(Unicode(1024))

    path = Column(Unicode(255), nullable=True)
    depth = Column(Integer)

    @property
    def descendants(self):
        path = '~%s' % self._id if self.path is None else self.path
        return app_model.DBSession.query(Category).filter(Category.path.like(path + '~%')).all()

    @property
    def children(self):
        try:
            next_depth = self.depth + 1
        except AttributeError:  # pragma: no cover
            # when self.depth is None, for migrated categories before version 0.4.0
            next_depth = 1
        return app_model.DBSession.query(Category).filter(Category.path.like(self.path + '~%'), Category.depth == next_depth).all()

    @property
    def parent(self):
        try:
            return app_model.DBSession.query(Category).filter(Category._id == self.path.split('~')[-1]).one()
        except NoResultFound:
            return None

    @property
    def siblings(self):
        return app_model.DBSession.query(Category).filter(Category.path == self.path, Category._id != self._id).all()

    @classmethod
    def by_id(cls, _id):
        return app_model.DBSession.query(cls).filter(cls._id == _id).one()


class CategoryImage(DeclarativeBase):
    __tablename__ = 'tgappcategories_images'

    _id = Column(Integer, autoincrement=True, primary_key=True)

    content = Column(UploadedFileField(upload_storage='category_image'))

    image_name = Column(Unicode(255))

    category_id = Column(Integer, ForeignKey('tgappcategories_categories._id'))
    category = relationship('Category', backref='images')



from os import getcwd

from webtest import TestApp
import transaction

from tg import AppConfig
from tg.configuration import milestones
from tg.configuration.auth import TGAuthMetadata
from tgext.pluggable import plug, app_model

from sqlalchemy import Integer, Column, Unicode, inspect
from sqlalchemy.orm import sessionmaker, scoped_session
from sqlalchemy.ext.declarative import declarative_base
from zope.sqlalchemy import ZopeTransactionExtension

import ming
from ming import Session
from ming.odm import ThreadLocalODMSession
from ming.odm.declarative import MappedClass
from ming.odm import FieldProperty
from ming import schema as s

from depot.manager import DepotManager


class FakeAppPackage(object):
    __file__ = __file__
    __name__ = 'tests'

    class lib(object):
        class helpers(object):
            pass
        helpers = helpers()

        class app_globals(object):
            class Globals():
                pass
        app_globals = app_globals()

    class websetup(object):
        def bootstrap(*args, **kwargs):
            pass


class FakeSQLAModel(object):
    def __init__(self):
        self.DeclarativeBase = declarative_base()
        self.DBSession = scoped_session(sessionmaker(autoflush=True, autocommit=False,
                                                     extension=ZopeTransactionExtension()))

        class User(self.DeclarativeBase):
            __tablename__ = 'tg_user'

            user_id = Column(Integer, autoincrement=True, primary_key=True)
            user_name = Column(Unicode(16), unique=True, nullable=False)
            email_address = Column(Unicode(255), unique=True, nullable=False)
            display_name = Column(Unicode(255))
            password = Column(Unicode(255), nullable=False)

        self.User = User

    def init_model(self, engine):
        self.DBSession.configure(bind=engine)
        self.DeclarativeBase.metadata.drop_all(engine)
        self.DeclarativeBase.metadata.create_all(bind=engine)


class FakeMingModel(object):
    def __init__(self):
        self.ming_session = Session()
        self.DBSession = ThreadLocalODMSession(self.ming_session)

        class User(MappedClass):
            class __mongometa__:
                session = self.DBSession
                name = 'tg_user'
                unique_indexes = [('user_name',), ('email_address',)]

            _id = FieldProperty(s.ObjectId)
            user_name = FieldProperty(s.String)
            email_address = FieldProperty(s.String)
            display_name = FieldProperty(s.String)
            password = FieldProperty(s.String)

        self.User = User

    def init_model(self, datastore):
        self.ming_session.bind = datastore

        try:
            # On MIM drop all data
            datastore.conn.drop_all()
        except TypeError:
            # On MongoDB drop database
            datastore.conn.drop_database(datastore.db)

        ming.odm.Mapper.compile_all()
        for mapper in ming.odm.Mapper.all_mappers():
            self.ming_session.ensure_indexes(mapper.collection)


class FakeUser(object):
    """
    Fake user that emulates an users without the need to actually
    query it from the database, it is able to trick sprox when
    resolving relations to the blog post Author.
    """
    def __int__(self):
        return 1

    def __getattr__(self, item):
        if item == 'user_id':
            return 1
        elif item == '_id':
            return self
        return super(FakeUser, self).__getattr__(item)


class TestAuthMetadata(TGAuthMetadata):
    def authenticate(self, environ, identity):
        return 'user'

    def get_user(self, identity, userid):
        if userid:
            return FakeUser()

    def get_groups(self, identity, userid):
        if userid:
            return ['managers']
        return []

    def get_permissions(self, identity, userid):
        if userid:
            return ['tgappcategories']
        return []

def configure_app(using):
    # Simulate starting configuration process from scratch
    milestones._reset_all()

    app_cfg = AppConfig(minimal=True)
    app_cfg.renderers = ['kajiki']
    app_cfg.default_renderer = 'kajiki'
    app_cfg.use_dotted_templatenames = True
    app_cfg.package = FakeAppPackage()
    app_cfg.use_toscawidgets2 = True
    app_cfg['tw2.enabled'] = True
    app_cfg.sa_auth.authmetadata = TestAuthMetadata()
    app_cfg['beaker.session.secret'] = app_cfg['session.secret'] = 'SECRET'
    app_cfg.auth_backend = 'ming'


    if using == 'sqlalchemy':
        app_cfg.package.model = FakeSQLAModel()
        app_cfg.use_sqlalchemy = True
        app_cfg['sqlalchemy.url'] = 'sqlite://'
        app_cfg.use_transaction_manager = True
        app_cfg['tm.enabled'] = True
        app_cfg.SQLASession = app_cfg.package.model.DBSession
    elif using == 'ming':
        app_cfg.package.model = FakeMingModel()
        app_cfg.use_ming = True
        app_cfg['ming.url'] = 'mim:///testcategories'
        app_cfg.MingSession = app_cfg.package.model.DBSession
    else:
        raise ValueError('Unsupported backend')

    app_cfg.model = app_cfg.package.model
    app_cfg.DBSession = app_cfg.package.model.DBSession

    if using == 'ming':  # ugly fix: depot can be configured just once
        # that if is just wrong, if sqlalchemy tests starts before it doesn't work
        storages = {
            'category_images': 'category_image',
        }
        for storage in storages:
            prefix = 'depot.%s.' % storage
            print('Configuring Storage %s*' % prefix)
            DepotManager.configure(storage, {
                'depot_backend_type': 'depot.io.memory.MemoryFileStorage',
                'depot.category_images.backend': 'depot.io.memory.MemoryFileStorage',
                'depot.category_images.prefix': 'category_images/'
            }, prefix)
            DepotManager.alias(storages[storage], storage)

    plug(app_cfg, 'tgappcategories', plug_bootstrap=False)
    return app_cfg


def create_app(app_config, auth=False):
    from tgappcategories import lib

    app = app_config.make_wsgi_app(skip_authentication=True)

    DepotManager._middleware = None
    app = DepotManager.make_middleware(app)

    if auth:
        app = TestApp(app, extra_environ=dict(REMOTE_USER='user'), relative_to=getcwd())
    else:
        app = TestApp(app)
        
    app.get('/non_existing_url_force_app_config_update', status=404)
    return app

def flush_db_changes():
    app_model.DBSession.flush()
    transaction.commit()



from tg.i18n import lazy_ugettext as l_
from tw2.core import Required, Deferred
from tw2.forms.widgets import Form, TextField, TextArea, SubmitButton, FileField, HiddenField, SingleSelectField, BaseLayout
from tgappcategories import model
from tg import config


pluggable_config = config['_pluggable_tgappcategories_config']


class FormLayout(BaseLayout):
    inline_engine_name = 'kajiki'
    template = '''
<div py:strip="True">
    <py:for each="c in w.children_hidden">
        ${c.display()}
    </py:for>
    <div>
        <span id="${w.compound_id}:error" class="error" >
            <p py:for="error in w.rollup_errors" class="alert alert-danger">
               <span class="glyphicon glyphicon-exclamation-sign"></span>
               ${error}
           </p>
        </span>
    </div>
    <div py:for="c in w.children_non_hidden"
         class="form-group ${((c.validator and getattr(c.validator, 'required', getattr(c.validator, 'not_empty', False))) and ' required' or '') + (c.error_msg and ' has-error' or '')}">
        <div class="col-sm-2">
            ${c.label}
        </div>
        <div class="col-sm-10">
            ${c.display()}
            <span id="${c.compound_id}:error" class="error help-block" py:content="c.error_msg"/>
        </div>
    </div>
</div>'''


class NewCategory(Form):    
    class child(FormLayout):
        image_small_id = HiddenField()
        image_big_id = HiddenField()

        name = TextField(label=l_('Name'), css_class='form-control',
                         validator=Required)

        description = TextArea(label=l_('Description'), rows=10, css_class='form-control',
                               validator=Required)

        image_small = FileField(label=pluggable_config.get('image1_label', l_('Small Image')),
                                css_class='form-control', attrs=dict(accept='image/*'))
        image_big = FileField(label=pluggable_config.get('image2_label', l_('Big Image')),
                              css_class='form-control', attrs=dict(accept='image/*'))
        parent_id = SingleSelectField(css_class='form-control', options=Deferred(lambda: [(c._id, c.name) for c in model.provider.query(model.Category, filters={})[1]]))
    submit = SubmitButton(css_class='btn btn-primary pull-right', value=l_('Create'))


class EditCategory(NewCategory):
    submit = SubmitButton(css_class='btn btn-primary pull-right', value=l_('Edit'))



# -*- coding: utf-8 -*-
import sys, os

try:
    from setuptools import setup, find_packages
except ImportError:
    from ez_setup import use_setuptools
    use_setuptools()
    from setuptools import setup, find_packages

install_requires = [
    "TurboGears2 >= 2.3.9",
    "tgext.pluggable >= 0.7.1",
    'sprox',
    "filedepot",
    'kajiki',
]

testpkgs = [
    'WebTest >= 1.2.3',
    'nose',
    'coverage',
    'ming',
    'sqlalchemy',
    'zope.sqlalchemy',
    'repoze.who',
    'tw2.forms',
]

here = os.path.abspath(os.path.dirname(__file__))
try:
    README = open(os.path.join(here, 'README.rst')).read()
except IOError:
    README = ''

setup(
    name='tgapp-categories',
    version='0.4.1',
    description='categories management system for web applications in turbogears2',
    long_description=README,
    author='Axant, Vincenzo Castilgia',
    author_email='vincenzo.castiglia@axant.it',
    url='https://github.com/axant/tgapp-categories',
    keywords='turbogears2.application',
    packages=find_packages(exclude=['ez_setup']),
    install_requires=install_requires,
    include_package_data=True,
    package_data={'tgappcategories': [
        'i18n/*/LC_MESSAGES/*.mo',
        'templates/*/*',
        'public/*/*'
    ]},
    message_extractors={'tgappcategories': [
            ('**.py', 'python', None),
            ('templates/**.xhtml', 'kajiki', None),
            ('public/**', 'ignore', None)
    ]},
    entry_points="""
    """,
    zip_safe=False,
    extras_require={
           'testing': testpkgs,
    },
)



import os
import re

import tg
from tgext.pluggable import app_model, instance_primary_key
from .base import configure_app, create_app, flush_db_changes
from tgappcategories import model
from webtest import AppError

from tgappcategories.helpers import images_with_image_name

find_urls = re.compile('http[s]?://(?:[a-zA-Z]|[0-9]|[$-_@.&+]|[!*\(\),]|(?:%[0-9a-fA-F][0-9a-fA-F]))+')
HERE = os.path.dirname(__file__)


class RegistrationControllerTests(object):
    def setup(self):
        self.app = create_app(self.app_config, False)

    def test_index(self):
        resp = self.app.get('/')
        assert 'HELLO' in resp.text

    def test_tgappcategories_index(self):
        resp = self.app.get('/tgappcategories', extra_environ={'REMOTE_USER': 'manager'})

        assert '/new_category' in resp.text, resp

    def test_tgappcategories_auth(self):
        try:
            self.app.get('/tgappcategories')
        except AppError as e:
            assert '401' in str(e)

    def test_create_category(self):
        self.app.post(
            '/tgappcategories/create_category',
            params={'name': 'category one', 'description': 'pretty description'},
            upload_files=[
                ('image_small',
                 os.path.join(HERE, '..', 'tgappcategories/public/images/star.png'))
            ],
            extra_environ={'REMOTE_USER': 'manager'},
            status=302,
        )
        __, cats = model.provider.query(model.Category,
                                        filters=dict(name='category one'))
        cat = cats[0]

        assert 'category one' == cat.name, cat.name
        assert 'pretty description' == cat.description, cat.description
        assert next(i for i in cat.images if i.image_name == 'image_small').content.url.startswith('/depot/category_images/')
        assert '' == cat.path, cat.path
        assert cat.depth == 1, cat.depth

    def test_create_subcategory(self):
        self.app.post(
            '/tgappcategories/create_category',
            params={'name': 'category one', 'description': 'pretty description'},
            extra_environ={'REMOTE_USER': 'manager'},
            status=302,
        )
        __, cats = model.provider.query(model.Category,
                                        filters=dict(name='category one'))
        cat = cats[0]
        self.app.post(
            '/tgappcategories/create_category',
            params={'name': 'category one.one', 'description': '1.1', 'parent_id': cat._id},
            extra_environ={'REMOTE_USER': 'manager'},
            status=302,
        )
        __, cats = model.provider.query(model.Category,
                                        filters=dict(name='category one.one'))
        subcat = cats[0]
        assert subcat.parent._id == cat._id
        assert subcat.path == '~%s' % subcat.parent._id, subcat.path
        assert subcat.depth == 2, subcat.depth

    def test_update_category(self):
        self.app.post(
            '/tgappcategories/create_category',
            params={'name': 'category one', 'description': 'pretty description'},
            upload_files=[
                ('image_small',
                 os.path.join(HERE, '..', 'tgappcategories/public/images/star.png'))
            ],
            extra_environ={'REMOTE_USER': 'manager'},
            status=302,
        )
        # Note that images are always created but with content = None
        __, cats = model.provider.query(model.Category, filters=dict(name='category one'))
        small_id = images_with_image_name(cats[0], 'image_small')[0]._id
        big_id = images_with_image_name(cats[0], 'image_big')[0]._id
        self.app.get(
            '/tgappcategories/update_category/' + str(cats[0]._id),
            params={'name': 'edited category', 'description': 'edited description',
                    'image_small_id': small_id, 'image_big_id': big_id},
            extra_environ={'REMOTE_USER': 'manager'},
            status=302,
        )
        __, cats = model.provider.query(model.Category,
                                        filters=dict(name='edited category'),
                                        )

        assert 'edited category' == cats[0].name, cats[0].name
        assert 'edited description' == cats[0].description, cats[0].description

    def test_move_subcategory_to_root(self):
        self.app.post(
            '/tgappcategories/create_category',
            params={'name': 'category one', 'description': 'pretty description'},
            extra_environ={'REMOTE_USER': 'manager'},
            status=302,
        )
        __, cats = model.provider.query(model.Category,
                                        filters=dict(name='category one'))
        cat = cats[0]
        self.app.post(
            '/tgappcategories/create_category',
            params={'name': 'category one.one', 'description': '1.1', 'parent_id': cat._id},
            extra_environ={'REMOTE_USER': 'manager'},
            status=302,
        )

        __, cats = model.provider.query(model.Category,
                                        filters=dict(name='category one.one'))
        moving = cats[0]
        moving.images  # force lazy loading
        self.app.post(
            '/tgappcategories/create_category',
            params={'name': 'category one.one.one', 'description': '1.1.1', 'parent_id': moving._id},
            extra_environ={'REMOTE_USER': 'manager'},
            status=302,
        )
        __, cats = model.provider.query(model.Category,
                                        filters=dict(name='category one.one.one'))
        child = cats[0]
        __, cats = model.provider.query(model.Category,
                                        filters=dict(name='category one'))
        root = cats[0]
        assert len(root.descendants) == 2
        assert len(root.children) == 1
        small_id = images_with_image_name(moving, 'image_small')[0]._id
        big_id = images_with_image_name(moving, 'image_big')[0]._id
        self.app.post(
            '/tgappcategories/update_category/' + str(moving._id),
            params={'name': 'category two', 'description': '2',
                    'parent_id': '',
                    'image_small_id': small_id, 'image_big_id': big_id},
            extra_environ={'REMOTE_USER': 'manager'},
            status=302,
        )
        __, cats = model.provider.query(model.Category,
                                        filters=dict(name='category two'))
        moved = cats[0]
        __, cats = model.provider.query(model.Category,
                                        filters=dict(name='category one.one.one'))
        child = cats[0]
        assert moved.path == '', moved.path
        assert moved.depth == 1
        assert len(moved.siblings) == 1
        assert child.path == '~%s' % moved._id, child.path
        assert child.depth == 2

    def test_move_category_with_descendants_to_another_parent(self):
        self.app.post(
            '/tgappcategories/create_category',
            params={'name': 'new parent', 'description': 'pretty description'},
            extra_environ={'REMOTE_USER': 'manager'},
            status=302,
        )
        __, cats = model.provider.query(model.Category,
                                        filters=dict(name='new parent'))

        new_parent = cats[0]
        self.app.post(
            '/tgappcategories/create_category',
            params={'name': 'old parent', 'description': 'pretty description'},
            extra_environ={'REMOTE_USER': 'manager'},
            status=302,
        )
        __, cats = model.provider.query(model.Category,
                                        filters=dict(name='old parent'))

        old_parent = cats[0]
        self.app.post(
            '/tgappcategories/create_category',
            params={'name': 'moving', 'description': 'pretty description', 'parent_id': old_parent._id},
            extra_environ={'REMOTE_USER': 'manager'},
            status=302,
        )

        __, cats = model.provider.query(model.Category,
                                        filters=dict(name='moving'))
        moving = cats[0]
        moving.images  # force lazy loading
        small_id = images_with_image_name(moving, 'image_small')[0]._id
        big_id = images_with_image_name(moving, 'image_big')[0]._id
        self.app.post(
            '/tgappcategories/create_category',
            params={'name': 'child', 'description': 'pretty description', 'parent_id': moving._id},
            extra_environ={'REMOTE_USER': 'manager'},
            status=302,
        )
        __, cats = model.provider.query(model.Category,
                                        filters=dict(name='child'))
        child = cats[0]
        self.app.post(
            '/tgappcategories/create_category',
            params={'name': 'descendant', 'description': 'pretty description', 'parent_id': child._id},
            extra_environ={'REMOTE_USER': 'manager'},
            status=302,
        )
        __, cats = model.provider.query(model.Category,
                                        filters=dict(name='descendant'))
        descendant = cats[0]
        assert len(old_parent.descendants) == 3
        assert len(old_parent.children) == 1
        self.app.post(
            '/tgappcategories/update_category/' + str(moving._id),
            params={'name': 'moved', 'description': '2',
                    'parent_id': str(new_parent._id),
                    'image_small_id': small_id, 'image_big_id': big_id},
            extra_environ={'REMOTE_USER': 'manager'},
            status=302,
        )
        __, cats = model.provider.query(model.Category,
                                        filters=dict(name='new parent'))

        new_parent = cats[0]
        __, cats = model.provider.query(model.Category,
                                        filters=dict(name='moved'))
        moved = cats[0]
        __, cats = model.provider.query(model.Category,
                                        filters=dict(name='child'))
        child = cats[0]
        __, cats = model.provider.query(model.Category,
                                        filters=dict(name='descendant'))
        descendant = cats[0]
        assert moved.path == '~%s' % new_parent._id, moved.path
        assert moved.depth == 2
        assert len(moved.siblings) == 0
        assert child.path == '~%s~%s' % (new_parent._id, moved._id), child.path
        assert child.depth == 3
        assert descendant.path == '~%s~%s~%s' % (new_parent._id, moved._id, child._id), descendant.path
        assert descendant.depth == 4

    def test_move_category_without_parent_to_a_parent(self):
        self.app.post(
            '/tgappcategories/create_category',
            params={'name': 'root', 'description': 'pretty description'},
            extra_environ={'REMOTE_USER': 'manager'},
            status=302,
        )
        __, cats = model.provider.query(model.Category,
                                        filters=dict(name='root'))

        root = cats[0]
        self.app.post(
            '/tgappcategories/create_category',
            params={'name': 'moving', 'description': 'pretty description'},
            extra_environ={'REMOTE_USER': 'manager'},
            status=302,
        )
        __, cats = model.provider.query(model.Category,
                                        filters=dict(name='moving'))
        moving = cats[0]
        moving.images  # force lazy loading
        small_id = images_with_image_name(moving, 'image_small')[0]._id
        big_id = images_with_image_name(moving, 'image_big')[0]._id
        self.app.post(
            '/tgappcategories/create_category',
            params={'name': 'child', 'description': 'pretty description', 'parent_id': moving._id},
            extra_environ={'REMOTE_USER': 'manager'},
            status=302,
        )
        __, cats = model.provider.query(model.Category,
                                        filters=dict(name='child'))
        child = cats[0]
        self.app.post(
            '/tgappcategories/create_category',
            params={'name': 'descendant', 'description': 'pretty description', 'parent_id': child._id},
            extra_environ={'REMOTE_USER': 'manager'},
            status=302,
        )
        __, cats = model.provider.query(model.Category,
                                        filters=dict(name='descendant'))
        descendant = cats[0]
        self.app.post(
            '/tgappcategories/update_category/' + str(moving._id),
            params={'name': 'moved', 'description': '2',
                    'parent_id': str(root._id),
                    'image_small_id': small_id, 'image_big_id': big_id},
            extra_environ={'REMOTE_USER': 'manager'},
            status=302,
        )
        __, cats = model.provider.query(model.Category,
                                        filters=dict(name='root'))

        root = cats[0]
        __, cats = model.provider.query(model.Category,
                                        filters=dict(name='moved'))
        moved = cats[0]
        __, cats = model.provider.query(model.Category,
                                        filters=dict(name='child'))
        child = cats[0]
        __, cats = model.provider.query(model.Category,
                                        filters=dict(name='descendant'))
        descendant = cats[0]
        assert moved.path == '~%s' % root._id, moved.path
        assert moved.depth == 2
        assert len(moved.siblings) == 0
        assert child.depth == 3
        assert child.path == '~%s~%s' % (root._id, moved._id), child.path
        assert descendant.depth == 4
        assert descendant.path == '~%s~%s~%s' % (root._id, moved._id, child._id), descendant.path

    def test_delete_category(self):
        self.app.get('/tgappcategories/create_category',
                     params={'name': 'category one', 'description': 'pretty description'},
                     extra_environ={'REMOTE_USER': 'manager'},
                     status=302,
                     )
        __, cats = model.provider.query(model.Category,
                                        filters=dict(name='category one'),
                                        )

        self.app.get('/tgappcategories/delete_category/' + str(cats[0]._id),
                     extra_environ={'REMOTE_USER': 'manager'},
                     status=302,
                     )

        count, cats = model.provider.query(model.Category,
                                           filters=dict(name='category one'),
                                           )
        assert count == 0, cats

    def test_cannot_delete_category(self):
        self.app.get('/tgappcategories/create_category',
                     params={'name': 'category one', 'description': 'pretty description'},
                     extra_environ={'REMOTE_USER': 'manager'},
                     status=302,
        )
        __, cats = model.provider.query(model.Category,
                                        filters=dict(name='category one'),
        )
        to_be_deleted = cats[0]
        self.app.post(
            '/tgappcategories/create_category',
            params={'name': 'category one.one', 'description': '1.1', 'parent_id': to_be_deleted._id},
            extra_environ={'REMOTE_USER': 'manager'},
            status=302,
        )
        
        self.app.get('/tgappcategories/delete_category/' + str(to_be_deleted._id),
                     extra_environ={'REMOTE_USER': 'manager'},
                     status=412,
        )
        count, cats = model.provider.query(model.Category,
                                           filters=dict(name='category one'),
        )
        assert count == 1, cats

    def test_new_category_form(self):
        resp = self.app.get('/tgappcategories/new_category',
                            extra_environ={'REMOTE_USER': 'manager'})

        assert 'name="name"' in resp.text, resp
        assert 'name="description"' in resp.text, resp
        assert '/create_category' in resp.text, resp

    def test_edit_category_form_404(self):
        self.app.get('/tgappcategories/edit_category',
                     extra_environ={'REMOTE_USER': 'manager'}, status=404)

    def test_edit_category_form(self):
        self.app.get('/tgappcategories/create_category',
                     params={'name': 'category one', 'description': 'pretty description'},
                     extra_environ={'REMOTE_USER': 'manager'},
                     status=302,
                     )
        __, cats = model.provider.query(model.Category,
                                        filters=dict(name='category one'),
                                        )

        resp = self.app.get('/tgappcategories/edit_category/' + str(cats[0]._id),
                            extra_environ={'REMOTE_USER': 'manager'}
                            )

        assert 'name="name"' in resp.text, resp
        assert 'name="description"' in resp.text, resp
        assert '/update_category' in resp.text, resp

    def test_edit_category_validator(self):
        self.app.get('/tgappcategories/create_category',
                     params={'name': 'category one', 'description': 'pretty description'},
                     extra_environ={'REMOTE_USER': 'manager'},
                     status=302,
                     )
        __, cats = model.provider.query(model.Category, filters=dict(name='category one'))

        resp = self.app.get('/tgappcategories/update_category/' + instance_primary_key(cats[0], True),
                            params={'name': '', 'description': ''},
                            extra_environ={'REMOTE_USER': 'manager'})
        assert 'id="name:error">Enter a value', resp


class TestCategoriesControllerSQLA(RegistrationControllerTests):
    @classmethod
    def setupClass(cls):
        cls.app_config = configure_app('sqlalchemy')


class TestCategoriesControllerMing(RegistrationControllerTests):
    @classmethod
    def setupClass(cls):
        cls.app_config = configure_app('ming')

